import os
import arcpy
from arcpy.sa import *
import requests
from datetime import datetime
import zipfile
import argparse

def create_rectangle_shapefile(lower_left, upper_right, output_shapefile):
    """
    Creates a shapefile with a rectangle defined by lower-left and upper-right coordinates.

    :param lower_left: Tuple of (longitude, latitude) for the lower-left corner.
    :param upper_right: Tuple of (longitude, latitude) for the upper-right corner.
    :param output_shapefile: Path to the output shapefile.
    """
    # Extract coordinates
    ll_lon, ll_lat = lower_left
    ur_lon, ur_lat = upper_right

    # Define rectangle coordinates (clockwise order)
    rectangle = arcpy.Polygon(arcpy.Array([
        arcpy.Point(ll_lon, ll_lat),
        arcpy.Point(ll_lon, ur_lat),
        arcpy.Point(ur_lon, ur_lat),
        arcpy.Point(ur_lon, ll_lat),
        arcpy.Point(ll_lon, ll_lat)  # Close the rectangle
    ]))

    # Create shapefile
    arcpy.env.overwriteOutput = True
    spatial_reference = arcpy.SpatialReference(4326)  # WGS 84
    arcpy.management.CreateFeatureclass(
        out_path=os.path.dirname(output_shapefile),
        out_name=os.path.basename(output_shapefile),
        geometry_type="POLYGON",
        spatial_reference=spatial_reference
    )

    # Add rectangle to shapefile
    with arcpy.da.InsertCursor(output_shapefile, ["SHAPE@"]) as cursor:
        cursor.insertRow([rectangle])

    print(f"Shapefile created at {output_shapefile}.")


def get_access_token(username: str, password: str) -> str:
    data = {
        "client_id": "cdse-public",
        "username": username,
        "password": password,
        "grant_type": "password",
    }
    try:
        r = requests.post(
            "https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
            data=data,
        )
        r.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Access token creation failed. Reponse from the server was: {r.json()}"
        )
    return r.json()["access_token"]


def get_refresh_token(username: str, password: str) -> str:
    data = {
        "client_id": "cdse-public",
        "username": username,
        "password": password,
        "grant_type": "password",
    }
    try:
        r = requests.post("https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
                          data=data,
                          )
        r.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Keycloak token creation failed. Reponse from the server was: {r.json()}"
        )
    return r.json()["refresh_token"]


def refresh_access_token(refresh_token: str) -> str:
    data = {
        "client_id": "cdse-public",
        "refresh_token": refresh_token,
        "grant_type": "refresh_token",
    }

    try:
        r = requests.post(
            "https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
            data=data,
        )
        r.raise_for_status()
    except Exception as e:
        raise Exception(
            f"Access token refresh failed. Reponse from the server was: {r.json()}"
        )

    return r.json()["access_token"]


def download_sentinel2_images_odata(area_shp, start_date, end_date, max_cloud_coverage, output_folder, username,
                                    password):
    """
    Downloads Sentinel-2 images for a specified area, time range, and cloud coverage using OData API.

    :param area_shp: Path to the shapefile defining the area of interest.
    :param start_date: Start date for the search (format: YYYY-MM-DD).
    :param end_date: End date for the search (format: YYYY-MM-DD).
    :param max_cloud_coverage: Maximum allowed cloud coverage (0-100).
    :param output_folder: Path to the folder where images will be saved.
    :param username: Username for the Copernicus account.
    :param password: Password for the Copernicus account.
    """
    # Generate API key
    api_key = get_access_token(username, password)

    # Ensure output folder exists
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Convert shapefile to GeoJSON for API query
    for row in arcpy.da.SearchCursor(area_shp, ["SHAPE@WKT"]):
        aoi = row[0]
        break

    # Define API endpoint and headers
    base_url = "https://catalogue.dataspace.copernicus.eu/odata/v1/Products"
    headers = {
        "Authorization": f"Bearer {api_key}",
    }

    data_collection = "SENTINEL-2"

    # Search for Sentinel-2 products
    search_url = f"{base_url}"
    params = {
        "$filter": (

            f"Collection/Name eq '{data_collection}' and OData.CSC.Intersects(area=geography'SRID=4326;{aoi}') "
            f"and ContentDate/Start gt {start_date}T00:00:00.000Z and ContentDate/Start lt {end_date}T00:00:00.000Z "
            f"and Attributes/OData.CSC.DoubleAttribute/any(att:att/Name eq 'cloudCover' and att/OData.CSC.DoubleAttribute/Value le {max_cloud_coverage}) "
            f"and Attributes/OData.CSC.StringAttribute/any(att:att/Name eq 'processingLevel' and att/OData.CSC.StringAttribute/Value eq 'S2MSI2A')"
        )
    }

    response = requests.get(search_url, headers=headers, params=params)
    response.raise_for_status()
    products = response.json().get('value', [])

    print(f"Total {len(products)} products to download.")

    # Download each product
    for product in products:
        product_id = product['Id']
        product_title = product['Name']

        print(f"Downloading {product_title}...")

        download_url = f"https://zipper.dataspace.copernicus.eu/odata/v1/Products({product_id})/$value"

        refresh_token = get_refresh_token(username, password)
        api_key = refresh_access_token(refresh_token)
        headers = {"Authorization": f"Bearer {api_key}", }

        product_response = requests.get(download_url, headers=headers, stream=True)
        product_response.raise_for_status()

        output_path = os.path.join(output_folder, f"{product_title}.zip")

        with open(output_path, 'wb') as file:
            for chunk in product_response.iter_content(chunk_size=8192):
                file.write(chunk)

    print(f"Downloaded {len(products)} products to {output_folder}.")


def find_band_name(strings, search_text):
    """
    Finds and returns the first string from a list that contains the specified search text.

    Parameters:
        strings (list of str): List of strings to search.
        search_text (str): Text to search for in the list.

    Returns:
        str: The first string containing the search text, or None if no match is found.
    """
    for string in strings:
        if search_text in string:
            return string
    return None


def set_band_descriptions(raster_path, descriptions):
    """
    Sets band descriptions for a multi-band raster.

    Parameters:
        raster_path (str): Path to the multi-band raster file.
        descriptions (list): List of descriptions for each band.
                             Length must match the number of bands.
    """
    # Get the number of bands in the raster
    raster = arcpy.Raster(raster_path)
    num_bands = raster.bandCount
    
    if len(descriptions) != num_bands:
        raise ValueError(f"Number of descriptions ({len(descriptions)}) does not match the number of bands ({num_bands}).")
    
    # Apply descriptions to each band
    old_names = raster.bandNames
    for i in range(0, num_bands):
        raster.renameBand(old_names[i], 'Band' + descriptions[i])
        
    print(f"Descriptions successfully set for raster: {raster_path}")
    

def process_sentinel2_to_tif(input_zip, output_tif, work_dir):
    """
    Processes Sentinel-2 data from a ZIP archive by extracting it, resampling all bands to the highest resolution,
    and combining them into a single TIFF file. Only bands in the IMG_DATA/R10m directory are processed.

    Parameters:
        input_zip (str): Path to the Sentinel-2 ZIP archive.
        output_tif (str): Path to the output TIFF file.
    """

    # Temporary folder for extraction
    temp_folder = os.path.join(work_dir, "sentinel2_temp")

    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)

    # Extract ZIP archive
    with zipfile.ZipFile(input_zip, 'r') as zip_ref:
        zip_ref.extractall(temp_folder)

    # Function to get raster resolution
    def get_raster_resolution(raster_path):
        desc = arcpy.Describe(raster_path)
        return desc.meanCellWidth, desc.meanCellHeight

    # Locate bands whose path contains 'IMG_DATA/R10m'
    bands = []
    for root, _, files in os.walk(temp_folder):
        for file in files:
            # print(os.path.join(root, file))
            if "IMG_DATA" in os.path.join(root, file) and \
               (file.endswith('_10m.jp2') or file.endswith('_20m.jp2') or file.endswith('_60m.jp2')):
                bands.append(os.path.join(root, file))

    if not bands:
        raise FileNotFoundError("No Sentinel-2 data (.jp2) found in IMG_DATA.")

    # Get the resolution of the band with the highest resolution (10 m Sentinel-2 bands)
    highest_resolution = float('inf')
    highest_resolution_band = None

    for band in bands:
        try:
            x_res, y_res = get_raster_resolution(band)
            resolution = max(x_res, y_res)
            if resolution < highest_resolution:
                highest_resolution = resolution
                highest_resolution_band = band
        except:
            continue

    # Reorder bands
    bands2export = ('_B01_20m', '_B02_10m', '_B03_10m', '_B04_10m', '_B05_20m', '_B06_20m',
                    '_B07_20m', '_B08_10m', '_B8A_20m', '_B09_60m', '_B11_20m', '_B12_20m') 
    

    # Resample all bands to the highest resolution
    resampled_bands = []
    for item in bands2export:
        band = find_band_name(bands, item)
        print(band)
        resampled_band = os.path.join(temp_folder, f"resampl{item}.tif")
        arcpy.Resample_management(band, resampled_band, highest_resolution, "BILINEAR")
        resampled_bands.append(resampled_band)

    # Combine bands into a single file
    arcpy.CompositeBands_management(resampled_bands, output_tif)

    # Set band descriptions
    set_band_descriptions(output_tif, bands2export)

    # Clean up temporary files
    for resampled_band in resampled_bands:
        if os.path.exists(resampled_band):
            os.remove(resampled_band)

    # Remove extracted files
    for root, _, files in os.walk(temp_folder, topdown=False):
        for file in files:
            os.remove(os.path.join(root, file))
        os.rmdir(root)

    print(f"Processing completed. Output file: {output_tif}")


def clip_raster_with_shapefile(input_raster, shapefile, output_raster):
    """
    Clips a raster to the area defined by a shapefile.

    Parameters:
        input_raster (str): Path to the input raster file.
        shapefile (str): Path to the shapefile defining the clipping area.
        output_raster (str): Path to the output clipped raster file.
    """
    # Perform the clipping
    arcpy.Clip_management(
        in_raster=input_raster,
        rectangle="#",
        out_raster=output_raster,
        in_template_dataset=shapefile,
        nodata_value="NONE",
        clipping_geometry="ClippingGeometry",
        maintain_clipping_extent="NO_MAINTAIN_EXTENT"
    )

    print(f"Clipping completed. Output file: {output_raster}")


def get_valid_zip_files(directory):
    """
    Returns a list of valid ZIP files in the specified directory.

    Parameters:
        directory (str): Path to the directory to scan for ZIP files.

    Returns:
        list: List of valid ZIP file paths.
    """
    if not os.path.isdir(directory):
        raise ValueError(f"The specified directory does not exist: {directory}")
    
    valid_zips = []
    for filename in os.listdir(directory):
        filepath = os.path.join(directory, filename)
        if filename.lower().endswith(".zip") and os.path.isfile(filepath):
            try:
                with zipfile.ZipFile(filepath, 'r') as zip_ref:
                    if zip_ref.testzip() is None:  # Check integrity of the ZIP file
                        valid_zips.append(filepath)
            except zipfile.BadZipFile:
                # Skip invalid ZIP files
                pass
    
    return valid_zips


def validate_and_parse_arguments():
    """
    Validates and parses command-line arguments for the script.
    Returns:
        dict: Dictionary of validated parameters.
    """
    parser = argparse.ArgumentParser(description="Download satellite data based on input parameters.")
    
    # Define the arguments
    parser.add_argument("-area", required=True, help="Geographic area as 'lon_min,lat_min,lon_max,lat_max'")
    parser.add_argument("-start", required=True, help="Start date in YYYYMMDD format")
    parser.add_argument("-end", required=True, help="End date in YYYYMMDD format")
    parser.add_argument("-max_cloud", type=int, required=True, help="Maximum cloud cover percentage (0-100)")
    parser.add_argument("-out_dir", required=True, help="Output directory path")
    parser.add_argument("-work_dir", required=True, help="Working directory path")
    parser.add_argument("-user", required=True, help="User name for authentication")
    parser.add_argument("-password", required=True, help="Password for authentication")
    
    # Parse arguments
    args = parser.parse_args()
    
    # Validate area
    try:
        area_coords = list(map(float, args.area.split(",")))
        if len(area_coords) != 4:
            raise ValueError
    except ValueError:
        raise ValueError("Invalid area format. Must be 'lon_min,lat_min,lon_max,lat_max' with four numbers.")
    
    # Validate start and end dates
    try:
        start_date = datetime.strptime(args.start, "%Y%m%d")
        end_date = datetime.strptime(args.end, "%Y%m%d")
        if start_date > end_date:
            raise ValueError("Start date must be before or equal to end date.")
    except ValueError as e:
        raise ValueError(f"Invalid date format: {e}")
    
    # Validate max_cloud
    if not (0 <= args.max_cloud <= 100):
        raise ValueError("max_cloud must be an integer between 0 and 100.")
    
    # Validate directories
    if not os.path.isdir(args.out_dir):
        raise ValueError(f"Output directory does not exist: {args.out_dir}")
    if not os.path.isdir(args.work_dir):
        raise ValueError(f"Working directory does not exist: {args.work_dir}")
    
    # Return validated arguments as a dictionary
    return {
        "area": area_coords,
        "start": start_date.strftime("%Y-%m-%d"),
        "end": end_date.strftime("%Y-%m-%d"),
        "max_cloud": args.max_cloud,
        "out_dir": os.path.abspath(args.out_dir),
        "work_dir": os.path.abspath(args.work_dir),
        "user": args.user,
        "password": args.password,
    }

# Main function
if __name__ == "__main__":

    # Environment settings
    arcpy.env.overwriteOutput = True
    
    try:
        params = validate_and_parse_arguments()

        # Reading command line parameters
        lower_left = (params['area'][0], params['area'][1])  # Longitude, Latitude
        upper_right = (params['area'][2], params['area'][3])  # Longitude, Latitude
        area_shp = os.path.join(params['work_dir'], "rectangle.shp")  # Output shapefile path
        start_date = params['start']  # Start date (YYYY-MM-DD)
        end_date = params['end']  # End date (YYYY-MM-DD)
        max_cloud_coverage = params['max_cloud']  # Max cloud coverage percentage
        output_folder = params['out_dir']  # Output folder for images
        work_dir = params['work_dir']  # Folder for temporary files
        user = params['user']  # Copernicus Open Access Hub username
        password = params['password']  # Copernicus Open Access Hub password

        # Example inputs
##        lower_left = (15.6247100, 50.6852986)  # Longitude, Latitude
##        upper_right = (15.7927664, 50.7487722)  # Longitude, Latitude
##        area_shp = "S:\\rectangle.shp"  # Output shapefile path
##        start_date = "20240801"  # Start date (YYYY-MM-DD)
##        end_date = "20240831"  # End date (YYYY-MM-DD)
##        max_cloud_coverage = 20  # Max cloud coverage percentage
##        output_folder = "S:\\images"  # Output folder for images
##        work_dir = "S:\\temp"  # Folder for temporary files
##        user = "<use your own username>"  # Copernicus Open Access Hub username
##        password = "<use your own password>"  # Copernicus Open Access Hub password

        create_rectangle_shapefile(lower_left, upper_right, area_shp)
        print('Downloading...')
        download_sentinel2_images_odata(area_shp, start_date, end_date, max_cloud_coverage, output_folder, user, password)
        image_list = get_valid_zip_files(output_folder)
        for image in image_list:
            print(image)
            print(' converting to tif...')
            out_image = image[:-3] + 'tif'
            process_sentinel2_to_tif(image, out_image, work_dir)
            print(' clipping...')
            clip_raster_with_shapefile(out_image, area_shp, out_image[:-4] + '_clip.tif')
    except RuntimeError as e:
        print(f"Error: {e}")









