import os
import arcpy
from arcpy.sa import *
import argparse

def parse_file(filename):
    """
    Reads a file and processes its lines into a list of lists.
    Each line is expected to contain two strings separated by a semicolon.

    Parameters:
        filename (str): The name of the file to read.

    Returns:
        list: A list of lists, where each sublist contains two strings.

    Raises:
        FileNotFoundError: If the file does not exist.
        ValueError: If a line does not contain exactly two strings separated by a semicolon.
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(f"The file '{filename}' does not exist.")

    result = {}

    with open(filename, 'r', encoding='utf-8') as file:
        for line_number, line in enumerate(file, start=1):
            line = line.strip()

            if line.startswith('#') or not line:
                continue

            parts = line.split(';')

            if len(parts) != 2:
                raise ValueError(
                    f"Invalid format on line {line_number}: '{line}'. Each line must contain exactly two parts separated by a semicolon."
                )

            result[parts[0].strip().lower()] = parts[1].strip()

    return result


def split_raster_bands(raster_path):
    """
    Reads a multi-band raster and stores each band in a separate variable named b_x,
    where x is the band number.

    Parameters:
        raster_path (str): Path to the raster file.

    Returns:
        dict: A dictionary where keys are band names (b_x) and values are arcpy.Raster objects.
    """
    try:
        # Load the raster
        raster = arcpy.Raster(raster_path)
        band_count = raster.bandCount

        # Ensure the raster has at least one band
        if band_count < 1:
            raise ValueError("The provided raster does not contain any bands.")

        # Store bands in a dictionary
        bands = {}
        for i in raster.bandNames:
            parts = i.split('_')

            if len(parts) != 3:
                raise ValueError(
                    f"Invalid band name {i} in file {raster_path}'."
                )

            band_variable = parts[1].lower()  # Variable name for the band
            bands[band_variable] = arcpy.Raster(raster_path + f"\\{i}")

        return bands

    except arcpy.ExecuteError:
        # Handle arcpy-specific errors
        raise RuntimeError(f"Error with arcpy tools: {arcpy.GetMessages()}")
    except Exception as e:
        # Handle general exceptions
        raise RuntimeError(f"An unexpected error occurred: {e}")


def validate_and_parse_arguments():
    """
    Validates and parses command-line arguments for the script.

    Returns:
        dict: Dictionary of validated parameters.
    """
    parser = argparse.ArgumentParser(description="Process raster file and calculate specified indices.")

    # Define command-line arguments
    parser.add_argument("-i", "--input_file", required=True, help="Path to the input raster file.")
    parser.add_argument("-idx_def", "--index_definition", required=True, help="Path to the file defining indices.")
    parser.add_argument("-idx", "--indices", required=True, help="Comma-separated list of indices (e.g., 'NDVI,NDBI').")
    parser.add_argument("-out_dir", "--output_directory", required=True, help="Path to the output directory.")

    # Parse arguments
    args = parser.parse_args()

    # Validate input file
    if not os.path.isfile(args.input_file):
        raise ValueError(f"Input file does not exist: {args.input_file}")

    # Validate indices
    indices = args.indices.split(",")
    if not all(index.isalnum() for index in indices):
        raise ValueError(f"Indices must be alphanumeric and comma-separated without spaces: {args.indices}")

    # Validate output directory
    if not os.path.isdir(args.output_directory):
        raise ValueError(f"Output directory does not exist: {args.output_directory}")

    # Validate index definition file, if provided
    if args.index_definition and not os.path.isfile(args.index_definition):
        raise ValueError(f"Index definition file does not exist: {args.index_definition}")

    # Return validated arguments as a dictionary
    return {
        "input_file": os.path.abspath(args.input_file),
        "indices": indices,
        "output_directory": os.path.abspath(args.output_directory),
        "index_definition": os.path.abspath(args.index_definition) if args.index_definition else None,
    }

# Main function
if __name__ == "__main__":

    # Environment settings
    arcpy.env.overwriteOutput = True

    try:
        params = validate_and_parse_arguments()

        in_raster = params['input_file']
        index_list_file = params['index_definition']
        process_index = params['indices']
        out_dir = params['output_directory']

        # in_raster = r"s:\temp\S2A_MSIL2A_20230907T100031_N0509_R122_T33UWS_20230907T173654.SAFE_clip.tif"
        # index_list_file = "veg_index_sentinel2.txt"
        # process_index = ['ari', 'atsavi', 'ndvi']
        # out_dir = 'S:\\indices'

        in_raster_base = os.path.basename(in_raster)

        # Read configuration
        all_index_dict = parse_file(index_list_file)

        # Read input file
        bands = split_raster_bands(in_raster)

        # Create variables
        b01 = bands['b01']
        b02 = bands['b02']
        b03 = bands['b03']
        b04 = bands['b04']
        b05 = bands['b05']
        b06 = bands['b06']
        b07 = bands['b07']
        b08 = bands['b08']
        b8a = bands['b8a']
        b09 = bands['b09']
        b11 = bands['b11']
        b12 = bands['b12']

        for idx_name in process_index:
            if idx_name.lower() not in all_index_dict.keys():
                print(f"Index '{idx_name}' not found in '{index_list_file}'. Its computation skipped.")
                continue

            # Compute index using eval()
            idx_expression = all_index_dict[idx_name.lower()].lower()
            try:
                result = eval(idx_expression, bands)
            except NameError:
                print(f"Syntax error in '{idx_name}' definition in '{index_list_file}'. Its computation skipped.")
                continue

            output_name = f"{in_raster_base[:-4]}_{idx_name}.tif"
            result.save(os.path.join(out_dir, output_name))
            print(f"Index {idx_name} computed successfully.")

    except RuntimeError as e:
        print(f"Error: {e}")

