import math
import os
import random
import arcpy
import argparse
import ast

import numpy as np
from arcpy.sa import *


def get_train_valid_count(in_all_data, in_prob, in_ratio, in_special):
    """Gets count of training and validation features from given input, considering share of training data
    in_ratio (0 < in_ratio < 1) and apriori probabilities of the classes. For classes labeled as special
    (typically very rare ones, where is insufficient count of terrain data available) will be in_ratio
    ignored, i.e. their share in validation data CAN be underestimated

    :param in_all_data: feature class of field data (i.e. mix of future training and validation data)
    :param in_prob: dict {classname => probability, ...}; sum of all probabilities is 1
    :param in_ratio: share of training data
    :param in_special: list of names of classes to behave in special manner
    :return: dict of the same structure as in_prob
    """
    all_total = 0
    all_cnt = {}

    cur = arcpy.da.SearchCursor(in_all_data, ('Classname',))
    for row in cur:
        all_total += 1
        if row[0] not in all_cnt.keys():
            all_cnt[row[0]] = 1
        else:
            all_cnt[row[0]] += 1

    # na trenovani vezmi vzdy prislusny podil vzorku dane tridy bez ohledu na to, jak je to z hlediska pomeru blbe
    # nasbirane
    train_cnt = {}
    train_total = 0
    for i in all_cnt.keys():
        train_cnt[i] = max(1, math.floor(all_cnt[i] * in_ratio))
        # musi byt aspon 1
        train_total += train_cnt[i]

    # validacni data jsou pak zbytek a mela by byt v prislusnem pomeru danem apriorni pravdepodobnosti vyskytu
    valid_total = all_total - train_total
    valid_teor_cnt = {}
    valid_avail_cnt = {}
    for i in all_cnt.keys():
        valid_teor_cnt[i] = math.floor(in_prob[i] * valid_total)
        valid_avail_cnt[i] = all_cnt[i] - train_cnt[i]

    # najdeme omezujici tridu
    # to je takova, ktera ma v potencialnich validacnich datech mensi zastoupeni, nez je apriorni pravdepodobnost
    # zaroven neni vzacna (abychom kvuli ni nemuseli omezovat hojne zastoupene tridy)
    # a zaroven se ji nejvic nedostava
    max_diff = 0
    key_class = None

    for i in all_cnt.keys():
        if valid_avail_cnt[i] >= valid_teor_cnt[i]:
            # je toho dost
            continue

        if i in in_special:
            # je vzacna
            continue

        diff = float(valid_teor_cnt[i] - valid_avail_cnt[i]) / valid_avail_cnt[i]
        if diff > max_diff:
            max_diff = diff
            key_class = i

    scale_factor = valid_avail_cnt[key_class] / valid_teor_cnt[key_class]
    # timto pomerem musime vsechny tridy mimo vzacnych prenasobit

    valid_real_cnt = {}

    for i in all_cnt.keys():
        if i not in in_special:
            valid_real_cnt[i] = math.floor(valid_teor_cnt[i] * scale_factor)
        else:
            if valid_avail_cnt[i] < valid_teor_cnt[i]:
                # je jich malo, pouzij vsechny
                valid_real_cnt[i] = valid_avail_cnt[i]
            else:
                # je jich dost, zpracuj to jako ostatni
                valid_real_cnt[i] = math.floor(valid_teor_cnt[i] * scale_factor)

    return (train_cnt, valid_real_cnt)


def prepare_train_valid_data(in_all_data, in_train_cnt, in_valid_cnt, out_train_data, out_valid_data):
    """Prepares training and validation data as random selection from field data with limiting conditions given by the
    parameters.

    :param in_all_data: feature class with field data
    :param in_train_cnt: dict {classname => count, ...} count of training data to be selected
    :param in_valid_cnt: dict {classname => count, ...} count of validation data to be selected
    :param out_train_data: feature class with training data
    :param out_valid_data: feature class with validation data
    :return: None
    """
    temp_dataset = "tmp_all_data.shp"
    arcpy.CopyFeatures_management(in_all_data, temp_dataset)
    arcpy.AddField_management(temp_dataset, 'FLAG', 'SHORT')

    r = arcpy.GetCount_management(temp_dataset)
    total_count = int(r.getOutput(0))

    random_order = list(range(0, total_count))
    random.shuffle(random_order)

    data = {}
    cur = arcpy.da.SearchCursor(temp_dataset, ('FID', 'Classname'))
    for row in cur:
        data[row[0]] = row[1]
    del cur

    flag = {}
    used = {}
    for i in in_train_cnt.keys():
        used[i] = [0, 0]

    for oid in random_order:
        if used[data[oid]][0] < in_train_cnt[data[oid]]:
            # je jeste prostor pro trenovaci data
            flag[oid] = 1  # bude to trenovaci dato
            used[data[oid]][0] += 1

        elif used[data[oid]][1] < in_valid_cnt[data[oid]]:
            flag[oid] = 2  # bude to validacni dato
            used[data[oid]][1] += 1

    cur = arcpy.da.UpdateCursor(temp_dataset, ('FID', 'FLAG'))
    for row in cur:
        if row[0] in flag.keys():
            row[1] = flag[row[0]]
        else:
            row[1] = 0

        cur.updateRow(row)
    del cur

    arcpy.MakeFeatureLayer_management(temp_dataset, 'tmp')
    arcpy.SelectLayerByAttribute_management('tmp', 'NEW_SELECTION', 'FLAG = 1')
    arcpy.CopyFeatures_management('tmp', out_train_data)

    arcpy.SelectLayerByAttribute_management('tmp', 'NEW_SELECTION', 'FLAG = 2')
    arcpy.CopyFeatures_management('tmp', out_valid_data)
    arcpy.Delete_management('tmp')


def get_accur_dict(dbf_file):
    """
    Extracts data from the second-to-last row and second-to-last column of a DBF file.

    Parameters:
        dbf_file (str): Path to the DBF file.

    Returns:
        tuple: A tuple containing:
            - A list of values from the second-to-last row.
            - A list of values from the second-to-last column.

    Raises:
        FileNotFoundError: If the DBF file does not exist.
        ValueError: If the DBF file has fewer than 2 rows or columns.
    """
    # Ensure the DBF file exists
    if not arcpy.Exists(dbf_file):
        raise FileNotFoundError(f"The DBF file {dbf_file} does not exist.")

    # Open the DBF file as a table
    with arcpy.da.SearchCursor(dbf_file, "*") as cursor:
        # Get all rows as a list
        rows = list(cursor)

        # Ensure there are at least two rows
        if len(rows) < 2:
            raise ValueError("The DBF file has fewer than two rows.")

        # Get the second-to-last row
        second_to_last_row = rows[-2]

    # Get field names and ensure there are at least two columns
    fields = [field.name for field in arcpy.ListFields(dbf_file)]
    if len(fields) < 2:
        raise ValueError("The DBF file has fewer than two columns.")

    # Get the second-to-last column values
    second_to_last_column_index = -2
    second_to_last_column = [
        row[second_to_last_column_index] for row in rows
    ]

    return {'ua': list(second_to_last_row[2:-3]), 'pa': second_to_last_column[:-3], 'oa': second_to_last_column[-2]}


def get_pixel_probability(raster_path):
    """
    Calculates the probability distribution of pixel values in a raster, excluding NoData values.

    Parameters:
        raster_path (str): Path to the input raster.

    Returns:
        dict: A dictionary where keys are pixel values, and values are their probabilities (occurrence divided by total valid pixels).
    """
    # Open the raster
    raster = arcpy.Raster(raster_path)

    # Read the raster as a NumPy array
    raster_array = arcpy.RasterToNumPyArray(raster, nodata_to_value=None)

    # Mask out NoData values
    valid_pixels = raster_array[raster_array != raster.noDataValue]

    # Get the total number of valid pixels
    total_valid_pixels = valid_pixels.size

    if total_valid_pixels == 0:
        raise ValueError("The raster contains only NoData values.")

    # Get the unique values and their counts
    unique, counts = np.unique(valid_pixels, return_counts=True)

    # Calculate probabilities
    probabilities = counts / total_valid_pixels

    # Create a dictionary from unique values and probabilities
    value_probabilities = dict(zip(unique, probabilities))

    return value_probabilities


def get_class_probabilities(pixel_stats, raster):
    class_names = {}
    out_prob = {}

    with arcpy.da.SearchCursor(raster, ["Class_name", "Value"]) as cursor:
        for row in cursor:
            name = row[0]
            value = row[1]
            class_names[name] = value

    for k in class_names.keys():
        out_prob[k] = pixel_stats[class_names[k]]

    return out_prob


def create_aligned_points_from_polygons(polygon_layer, raster_layer, output_point_layer):
    """
    Generates points at the center of raster cells within the area of polygons.
    Transfers polygon attributes to the points using a spatial join.

    Parameters:
        polygon_layer (str): Path to the input polygon feature layer.
        raster_layer (str): Path to the input raster layer.
        output_point_layer (str): Path to the output point feature layer.
    """
    # Set environment
    arcpy.env.snapRaster = raster_layer

    # Get raster properties to align the polygon raster
    raster_desc = arcpy.Raster(raster_layer)
    cell_size = raster_desc.meanCellWidth
    extent = raster_desc.extent
    spatial_ref = raster_desc.spatialReference

    # Create a temporary raster from the polygons aligned to the input raster
    temp_polygon_raster = arcpy.env.scratchGDB + "/polygon_raster"
    arcpy.PolygonToRaster_conversion(
        in_features=polygon_layer,
        value_field="Classname",
        out_rasterdataset=temp_polygon_raster,
        cell_assignment="MAXIMUM_AREA",
        priority_field="NONE",
        cellsize=cell_size,
    )

    # Convert the temporary polygon raster to points
    temp_points = arcpy.env.scratchGDB + "/polygon_points"
    arcpy.RasterToPoint_conversion(
        in_raster=temp_polygon_raster,
        out_point_features=temp_points,
        raster_field="VALUE"
    )

    # Perform a spatial join to transfer polygon attributes to the points
    arcpy.analysis.SpatialJoin(
        target_features=temp_points,
        join_features=polygon_layer,
        out_feature_class=output_point_layer,
        join_type="KEEP_COMMON",
        match_option="WITHIN"
    )

    # Clean up temporary files
    arcpy.management.Delete(temp_polygon_raster)
    arcpy.management.Delete(temp_points)
    cleanup_shapefile(output_point_layer, ["Classname", "Classvalue", "classname", "classvalue"])


def cleanup_shapefile(input_shapefile, keep_fields):
    """
    Removes all fields from a shapefile except FID, Shape, and specified fields.

    Parameters:
        input_shapefile (str): Path to the input shapefile.
        keep_fields (list): List of field names to retain in addition to FID and Shape.

    Returns:
        None
    """
    # Ensure the input shapefile exists
    if not arcpy.Exists(input_shapefile):
        raise FileNotFoundError(f"The shapefile {input_shapefile} does not exist.")

    # Get a list of all fields in the shapefile
    fields = arcpy.ListFields(input_shapefile)

    # Automatically keep FID and Shape fields
    default_keep_fields = ["FID", "Shape"]
    fields_to_keep = set(default_keep_fields + keep_fields)

    # Identify fields to delete
    fields_to_delete = [field.name for field in fields if field.name not in fields_to_keep]

    # Delete unwanted fields
    if fields_to_delete:
        arcpy.management.DeleteField(input_shapefile, fields_to_delete)


def validate_shapefile_attributes(input_shapefile):
    """
    Validates that the input shapefile contains the required attributes 'classname' and 'classvalue'.

    Parameters:
        input_shapefile (str): Path to the input shapefile.

    Raises:
        FileNotFoundError: If the shapefile does not exist.
        ValueError: If the required attributes are missing.
    """
    # Ensure the shapefile exists
    if not arcpy.Exists(input_shapefile):
        raise FileNotFoundError(f"The shapefile {input_shapefile} does not exist.")

    # Get a list of all field names in the shapefile
    field_names = [field.name.lower() for field in arcpy.ListFields(input_shapefile)]

    # Required fields
    required_fields = {"classname", "classvalue"}

    # Check for missing fields
    missing_fields = required_fields - set(field_names)
    if missing_fields:
        raise ValueError(f"The shapefile {input_shapefile} is missing required attributes: {', '.join(missing_fields)}")


def prepare_classif_fields(input_shapefile):
    """
    Adds the fields 'CLASSIFIED' and 'GRNDTRUTH' of type LONG to the input shapefile.
    If these fields already exist, they are deleted and recreated.

    Parameters:
        input_shapefile (str): Path to the input shapefile.

    Raises:
        FileNotFoundError: If the shapefile does not exist.
    """
    # Ensure the shapefile exists
    if not arcpy.Exists(input_shapefile):
        raise FileNotFoundError(f"The shapefile {input_shapefile} does not exist.")

    # Fields to add or replace
    fields_to_add = ["CLASSIFIED", "GRNDTRUTH"]

    # Get existing fields in the shapefile
    existing_fields = [field.name for field in arcpy.ListFields(input_shapefile)]

    # Delete fields if they exist
    for field in fields_to_add:
        if field in existing_fields:
            arcpy.management.DeleteField(input_shapefile, field)

    # Add new fields
    for field in fields_to_add:
        arcpy.management.AddField(input_shapefile, field_name=field, field_type="LONG")

    arcpy.CalculateField_management(input_shapefile, 'GRNDTRUTH', '!Classvalue!', 'PYTHON')


def primary_eliminate(in_raster, out_vector, limit, iters=2):
    """Performs primary elimination of undersized features in classified raster according to the given parametres.

    :param in_raster: raster with a result of classification
    :param out_vector: output feature class name
    :param limit: the minimum size of feature (in units of raster's coordinate system)
    :param iters: the maximum number of iterations to be performed
    :return: None
    """
    arcpy.RasterToPolygon_conversion(in_raster, 'tmpx_polygon_0.shp', 'NO_SIMPLIFY', 'Class_name',
                                     'SINGLE_OUTER_PART', None)

    cnt = 99999999
    for i in range(0, iters):
        result = arcpy.GetCount_management('tmpx_polygon_{}.shp'.format(i))
        last_cnt = cnt
        last_i = i
        cnt = int(result[0])
        print(f"  patches: {cnt}")  # pro informaci

        if cnt == last_cnt:
            # uz se nezlepsilo
            break

        arcpy.CalculateGeometryAttributes_management('tmpx_polygon_{}.shp'.format(i), [['Area', 'AREA']])
        arcpy.MakeFeatureLayer_management('tmpx_polygon_{}.shp'.format(i), 'tmp_sel')
        arcpy.SelectLayerByAttribute_management('tmp_sel', 'NEW_SELECTION', 'Area < {}'.format(limit))
        arcpy.Eliminate_management('tmp_sel', 'tmpx_polygon_{}.shp'.format(i + 1), 'AREA')
        arcpy.Delete_management('tmp_sel')

    arcpy.Dissolve_management('tmpx_polygon_{}.shp'.format(last_i), out_vector, ['gridcode', 'Class_name'])

    c = arcpy.ListFeatureClasses('tmpx_*')
    for i in c:
        arcpy.Delete_management(i)


def create_geodatabase_for_result(gdb_path, spatial_ref):
    """
    Creates a file geodatabase with a polygon feature class named 'result'.
    The feature class will have the specified coordinate system and the following fields:
    - 'ITER' (LONG)
    - 'UA' (DOUBLE)
    - 'PA' (DOUBLE)
    - 'F1SCORE' (DOUBLE)
    - 'CLASS_NAME' (TEXT)
    - 'CLASS_VALUE' (LONG)

    If the geodatabase already exists, it will be deleted and recreated.

    Parameters:
        gdb_path (str): Path where the geodatabase will be created.
        spatial_ref (arcpy.SpatialReference): Spatial reference for the feature class.

    Returns:
        str: Path to the created feature class.

    Raises:
        Exception: If the geodatabase cannot be created or feature class cannot be added.
    """
    # Ensure the output directory exists
    output_dir, gdb_name = os.path.split(gdb_path)
    if not arcpy.Exists(output_dir):
        raise FileNotFoundError(f"The output directory '{output_dir}' does not exist.")

    # Check if the geodatabase already exists
    if arcpy.Exists(gdb_path):
        print(f"Geodatabase '{gdb_path}' already exists. Deleting it.")
        arcpy.Delete_management(gdb_path)

    # Create the geodatabase
    arcpy.CreateFileGDB_management(output_dir, gdb_name)

    # Define the name of the feature class
    feature_class_name = "result"
    feature_class_path = os.path.join(gdb_path, feature_class_name)

    # Create the feature class
    arcpy.CreateFeatureclass_management(
        out_path=gdb_path,
        out_name=feature_class_name,
        geometry_type="POLYGON",
        spatial_reference=spatial_ref
    )

    # Add fields to the feature class
    fields = [
        ("ITER", "LONG"),
        ("UA", "DOUBLE"),
        ("PA", "DOUBLE"),
        ("F1SCORE", "DOUBLE"),
        ("CLASS_NAME", "TEXT", 50),  # 50-character limit for text field
        ("CLASS_VALUE", "LONG"),
    ]

    for field_name, field_type, *args in fields:
        arcpy.AddField_management(
            in_table=feature_class_path,
            field_name=field_name,
            field_type=field_type,
            field_length=args[0] if args else None
        )

    return feature_class_path


def save_result_to_db(in_vec, iter, in_stat, out_fc):
    """Save result of classification with metadata to a given feature class

    :param in_vec: name of feature class with classification
    :param iter: classification number (of multiple classifications)
    :param in_stat: dict of classification statistics
    :param out_fc: name of feature class to append the in_vec to
    :return: None
    """
    s_cur = arcpy.da.SearchCursor(in_vec, ['SHAPE@', 'gridcode', 'Class_name'])
    i_cur = arcpy.da.InsertCursor(out_fc, ['SHAPE@',  # 0
                                           'ITER',  # 1
                                           'UA', 'PA', 'F1SCORE',  # 2, 3, 4
                                           'CLASS_NAME', 'CLASS_VALUE'])  # 5, 6
    for s_row in s_cur:
        i_row = 7 * [None]
        i_row[0] = s_row[0]
        i_row[1] = iter

        i_row[2] = in_stat['ua'][s_row[1] - 1]
        i_row[3] = in_stat['pa'][s_row[1] - 1]
        i_row[4] = (2 * i_row[2] * i_row[3]) / (i_row[2] + i_row[3])

        i_row[5] = s_row[2]
        i_row[6] = s_row[1]

        i_cur.insertRow(i_row)

    del s_cur, i_cur


def create_max_value_raster(raster_list, value_list, output_raster_path):
    """
    Creates a raster where each pixel's value is the number corresponding to the raster
    with the maximum pixel value at that position, considering NoData as zero.

    Parameters:
        raster_list (list): List of input rasters (file paths).
        value_list (list): List of numeric values corresponding to each raster.
        output_raster_path (str): File path for the output raster.

    Returns:
        None: The function saves the resulting raster to the specified path.
    """
    try:
        # Check if the lengths of the lists match
        if len(raster_list) != len(value_list):
            raise ValueError("The length of the raster list and value list must be the same.")

        # Ensure all rasters are loaded and processed
        rasters = [Raster(r) for r in raster_list]

        # Create an empty list to hold conditional rasters
        conditional_rasters = []

        for idx, raster in enumerate(rasters):
            # Replace NoData with 0
            processed_raster = Con(IsNull(raster), 0, raster)
            # Create a conditional raster: if this raster is the maximum, assign its value from value_list
            condition = (processed_raster == CellStatistics(rasters, "MAXIMUM"))
            conditional_raster = Con(condition, value_list[idx])
            conditional_rasters.append(conditional_raster)

        # Combine all conditional rasters to create the final output raster
        final_raster = CellStatistics(conditional_rasters, "SUM")

        # Save the resulting raster
        final_raster.save(output_raster_path)
        print(f"Output raster saved to {output_raster_path}")

    except arcpy.ExecuteError:
        # Handle arcpy-specific errors
        raise RuntimeError(f"Error with arcpy tools: {arcpy.GetMessages()}")
    except Exception as e:
        # Handle general exceptions
        raise RuntimeError(f"An unexpected error occurred: {e}")


def select_top_features(feature_class, top):
    """
    Adds a SELECTED field to the input feature class and sets its value to 1
    for the top 'top' records with the highest F1SCORE within each CLASS_NAME.
    Other records will have a value of 0.

    Parameters:
        feature_class (str): Path to the input feature class in a geodatabase.
        top (int): The number of top records to select for each CLASS_NAME.

    Returns:
        None: The function modifies the input feature class in place.
    """
    try:
        # Check if the input feature class exists
        if not arcpy.Exists(feature_class):
            raise ValueError(f"The feature class '{feature_class}' does not exist.")

        # Add the SELECTED field if it doesn't already exist
        field_name = "SELECTED"
        if field_name not in [field.name for field in arcpy.ListFields(feature_class)]:
            arcpy.AddField_management(feature_class, field_name, "SHORT")

        # Initialize SELECTED field to 0 for all records
        with arcpy.da.UpdateCursor(feature_class, [field_name]) as cursor:
            for row in cursor:
                row[0] = 0
                cursor.updateRow(row)

        # Group by CLASS_NAME and select top records by SCORE
        with arcpy.da.SearchCursor(feature_class, ["OID@", "CLASS_NAME", "F1SCORE"],
                                   sql_clause=(None, "ORDER BY CLASS_NAME, F1SCORE DESC")) as cursor:
            current_class = None
            count = 0
            top_oids = []

            for oid, class_name, score in cursor:
                if class_name != current_class:
                    current_class = class_name
                    count = 1
                else:
                    count += 1

                if count <= top:
                    top_oids.append(oid)

        # Update the SELECTED field to 1 for the selected OIDs
        with arcpy.da.UpdateCursor(feature_class, ["OID@", field_name]) as cursor:
            for oid, selected in cursor:
                if oid in top_oids:
                    cursor.updateRow([oid, 1])

    except arcpy.ExecuteError:
        raise RuntimeError(f"Error with arcpy tools: {arcpy.GetMessages()}")
    except Exception as e:
        raise RuntimeError(f"An unexpected error occurred: {e}")


def map_raster_to_values(raster_list, mapping_dict):
    """
    Maps values from a list of raster filenames to corresponding values in a dictionary.

    Parameters:
        raster_list (list): A list of raster filenames in the format 'result_X.tif'.
        mapping_dict (dict): A dictionary where keys are X values from filenames, and values are corresponding A values.

    Returns:
        list: A list of A values corresponding to the X values in the filenames.
    """
    result = []
    for raster in raster_list:
        try:
            # Extract X from the filename (assumes the format 'result_X.tif')
            x_value = raster.split('_')[1].split('.')[0]

            # Map the X value to the corresponding A value
            if x_value in mapping_dict:
                result.append(mapping_dict[x_value])
            else:
                raise KeyError(f"No mapping found for X: {x_value}")
        except (IndexError, KeyError) as e:
            raise ValueError(f"Error processing filename '{raster}': {e}")

    return result


def validate_args():
    """
    Validates the input arguments for the classifier.py script and returns a dictionary of parameters.

    Returns:
        dict: Dictionary containing the validated input parameters.

    Raises:
        ValueError: If any parameter fails validation.
    """
    parser = argparse.ArgumentParser(
        description="Classifier script for raster processing using training and validation data."
    )

    # Required arguments
    parser.add_argument("-i", "--input_file", required=True, help="Input raster file.")
    parser.add_argument("-out", "--output_file", required=True, help="Output classified raster file.")
    parser.add_argument("-work", "--work_dir", required=True, help="Directory for temporary files.")
    parser.add_argument("-tv", "--training_validation_data", required=True, help="Training and validation data (shapefile).")

    # Optional arguments
    parser.add_argument("-ratio", "--tv_ratio", type=float, default=0.33, help="Training-validation split ratio (default: 0.33).")
    parser.add_argument("-aprob", "--in_prob", type=str, help="A priori probabilities (Python dictionary).")
    parser.add_argument("-special", "--special_classes", type=str, help="List of special classes (comma-separated).")
    parser.add_argument("-repeat", "--repeat_classification", type=int, default=15, help="Number of classifications to perform (default: 15).")
    parser.add_argument("-min_patch", "--pixels", type=int, default=5, help="Minimum patch size in pixels (default: 5).")
    parser.add_argument("-simpl", "--repeat_simpl", type=int, default=4, help="Maximum iterations for simplifying small patches (default: 4).")
    parser.add_argument("-select", "--select_for_result", type=int, default=5, help="Number of classifications to select for result (default: 5).")

    args = parser.parse_args()

    # Validate paths
    if not os.path.exists(args.input_file):
        raise ValueError(f"Input file does not exist: {args.input_file}")
    if not os.path.isdir(args.work_dir):
        raise ValueError(f"Work directory does not exist: {args.work_dir}")
    if not os.path.exists(args.training_validation_data):
        raise ValueError(f"Training-validation data does not exist: {args.training_validation_data}")

    # Validate a priori probabilities
    if args.in_prob:
        try:
            in_prob_dict = ast.literal_eval(args.in_prob)
            if not isinstance(in_prob_dict, dict):
                raise ValueError("A priori probabilities must be a dictionary.")
            total_prob = sum(in_prob_dict.values())
            if not 0.99 <= total_prob <= 1.01:
                raise ValueError("A priori probabilities must sum to 1.")
        except (SyntaxError, ValueError):
            raise ValueError("Invalid format for a priori probabilities. Must be a valid Python dictionary.")
    else:
        in_prob_dict = None

    # Validate special classes
    if args.special_classes:
        special_classes = args.special_classes.split(",")
    else:
        special_classes = []

    # Validate repeat_classification and select_for_result
    if args.select_for_result >= args.repeat_classification:
        raise ValueError("select_for_result must be less than repeat_classification.")

    # Prepare and return the parameters as a dictionary
    params = {
        "input_file": args.input_file,
        "output_file": args.output_file,
        "work_dir": args.work_dir,
        "training_validation_data": args.training_validation_data,
        "tv_ratio": args.tv_ratio,
        "in_prob": in_prob_dict,
        "special_classes": special_classes,
        "repeat_classification": args.repeat_classification,
        "pixels": args.pixels,
        "repeat_simpl": args.repeat_simpl,
        "select_for_result": args.select_for_result,
    }

    return params

if __name__ == "__main__":
    try:
        params = validate_args()

        base_dir = params['work_dir']

        in_image = params['input_file']

        in_tv_data = params['training_validation_data']
        in_training_validation_ratio = params['tv_ratio']

        in_prob = params['in_prob']
        in_special_classes = params['special_classes']
        # rare classes, special behaviour required

        repeat_classification = params['repeat_classification']

        min_result_patch_pixels = params['pixels']
        min_result_area = (arcpy.Raster(in_image).meanCellWidth ** 2) * min_result_patch_pixels
        max_iter_simpl = params['repeat_simpl']

        best_count_to_select = params['select_for_result']

        output_classification_raster = params['output_file']

        # base_dir = 'S:\\classification\\'
        # arcpy.env.workspace = base_dir
        # in_tv_data = 'tv_data_pnt_aligned.shp'
        # in_training_validation_ratio = 0.33
        #
        # in_image = 'S:\\classification\\2021_07_13_BL_MSSRGB_3cm_envi.tif'
        #
        # in_prob = {'afs': 0.09212, 'cv': 0.17971, 'cxbig': 0.02820, 'desch': 0.11637, 'mol': 0.17491, 'nard': 0.40632,
        #            'smrk': 0.00238}
        # in_prob = None
        # in_special_classes = ('cxbig', 'smrk', 'afs')
        # # rare classes, special behaviour required
        #
        #
        # repeat_classification = 3
        #
        # min_result_patch_pixels = 10
        # min_result_area = arcpy.Raster(in_image).meanCellWidth **2 * min_result_patch_pixels
        # max_iter_simpl = 4
        #
        # best_count_to_select = 2
        # output_classification_raster = 'S:\\classification\\2021_07_13_BL_MSSRGB_3cm_envi_classified.tif'

        arcpy.env.overwriteOutput = 1
        arcpy.env.workspace = base_dir

        # check TV data and convert polygons inputs to points if necessary
        print('Checking and preparing training/validation data...')
        validate_shapefile_attributes(in_tv_data)

        tv_data_final = 'tmp_tv_data_final.shp'

        desc = arcpy.Describe(in_tv_data)
        geometry_type = desc.shapeType.lower()

        # Check the geometry type and call the corresponding function
        if geometry_type == "point":
            arcpy.CopyFeatures_management(in_tv_data, tv_data_final)
        elif geometry_type == "polygon":
            create_aligned_points_from_polygons(in_tv_data, in_image, tv_data_final)
        else:
            raise ValueError(f"Unsupported geometry type for T/V data: {geometry_type}")

        prepare_classif_fields(tv_data_final)

        # a priori probabilities are not given, we have to estimate them
        if not in_prob:
            print('A priori probabilities are not given. Trying to estimate them...')
            # use all input data for training
            TrainMaximumLikelihoodClassifier(in_image, tv_data_final, 'def_preclass.ecd')
            # classify
            class_raster = ClassifyRaster(in_image, 'def_preclass.ecd')
            class_raster.save('tmp_preclass.tif')
            # compute statistics
            pixel_count = get_pixel_probability('tmp_preclass.tif')
            in_prob = get_class_probabilities(pixel_count, 'tmp_preclass.tif')

        (tc, vc) = get_train_valid_count(tv_data_final, in_prob, in_training_validation_ratio, in_special_classes)

        print('Preparing database for results...')
        result_database = os.path.join(base_dir, 'results.gdb')
        create_geodatabase_for_result(result_database, arcpy.Raster(in_image).spatialReference)

        max_classification = repeat_classification
        class_accur_info = {}

        for iter in range(0, max_classification):
            print(f"Classification: {iter + 1}/{max_classification}")
            # priprava trenovacich / validacnich dat
            print(' selecting training and validation data...')
            prepare_train_valid_data(tv_data_final, tc, vc, 'tmp_tp.shp', 'tmp_vp.shp')

            print(' classification...')
            if (iter % 3) == 0:
                method = 'MLC'
                TrainMaximumLikelihoodClassifier(in_image, 'tmp_tp.shp', 'def_1.ecd')
                # 0.21 # 0.86
            elif (iter % 3) == 1:
                method = 'SVM'
                TrainSupportVectorMachineClassifier(in_image, 'tmp_tp.shp', "def_1.ecd")
                # 0.73 # 0.88
            elif (iter % 3) == 2:
                method = 'RTR'
                TrainRandomTreesClassifier(in_image, 'tmp_tp.shp', "def_1.ecd", max_num_trees=100)
                # 0.654, 0.656 // 0.66 numtrees=100, numtrees=500 # 0.84

            class_image = in_image.replace('.tif', '_class_{}_{:03d}.tif'.format(method, iter))

            # klasifikace
            class_raster = ClassifyRaster(in_image, 'def_1.ecd')
            class_raster.save(class_image)

            # vyhodnoceni presnosti
            print(' accuracy assessment...')
            UpdateAccuracyAssessmentPoints(class_image, 'tmp_vp.shp', 'tmp_valid_final.shp', 'CLASSIFIED')

            ComputeConfusionMatrix("tmp_valid_final.shp", "tmp_conf_matrix.dbf")

            # save info about accuracy
            accur = get_accur_dict("tmp_conf_matrix.dbf")

            print(' result simplification...')
            primary_eliminate(class_image, f"tmp_result_{iter}.shp", min_result_area, max_iter_simpl)

            print(' saving result...')
            save_result_to_db(f"tmp_result_{iter}.shp", iter, accur, os.path.join(result_database, 'result'))

        print('Combining partial results...')

        # Select n-best classifications
        select_top_features(os.path.join(result_database, 'result'), best_count_to_select)

        # Rasterize and merge them together
        cur = arcpy.da.SearchCursor(os.path.join(result_database, 'result'),
                                    ['ITER', 'F1SCORE', 'CLASS_NAME', 'CLASS_VALUE', 'OBJECTID'], where_clause='SELECTED = 1')
        #                             0           1              2             3       4


        arcpy.env.snapRaster = in_image
        arcpy.env.extent = arcpy.Raster(in_image).extent
        pixel_size = arcpy.Raster(in_image).meanCellWidth

        labels = {}
        result = {}

        for row in cur:
            print('* ', end='', flush=True)
            iter = row[0]
            f1score = row[1]
            class_name = row[2]
            object_id = row[4]
            labels[class_name] = row[3]

            try:
                out_raster_name = f'tmp_class_{iter}_{class_name}.tif'

                if arcpy.Exists(out_raster_name):
                    r1 = arcpy.Raster(out_raster_name)
                else:
                    query = f'"OBJECTID" = {object_id}'
                    arcpy.Select_analysis(os.path.join(result_database, 'result'), "in_memory/temp", query)

                    arcpy.FeatureToRaster_conversion("in_memory/temp", "F1SCORE", 'tmp.tif', pixel_size)
                    r = arcpy.Raster('tmp.tif')
                    r1 = arcpy.sa.Con(arcpy.sa.IsNull(r), 0, r)
                    r1.save(out_raster_name)

                if class_name not in result.keys():
                    result[class_name] = {}
                    result[class_name]['r'] = r1
                    result[class_name]['cnt'] = 1
                else:
                    result[class_name]['r'] = result[class_name]['r'] + r1  # secteme
                    result[class_name]['cnt'] = result[class_name]['cnt'] + 1
            except:
                print('Error in aggregation: {}'.format(out_raster_name))

        # ulozit vysledne rastry:
        print('\nSaving:')
        for class_name in result.keys():
            try:
                print('~', end='', flush=True)
                r = result[class_name]['r']
                cnt = result[class_name]['cnt']
                output = r / float(cnt)
                output_name = f'result_{class_name}.tif'
                output.save(output_name)
            except:
                print('Error in aggregation: {}'.format(output_name))

        print('\n')

        raster_list = arcpy.ListRasters('result_*.tif')
        value_list = map_raster_to_values(raster_list, labels)
        create_max_value_raster(raster_list, value_list, output_classification_raster)

        print('Finished successfully!')

    except ValueError as e:
        print(f"Error: {e}")