import arcpy
import argparse
import os


def process_vector_layers(input_file, output_fc):
    """
    Processes vector layers from an input file, assigns a new attribute `T_x` based on specified attributes,
    and merges them into a single feature class.

    Parameters:
        input_file (str): Path to the input text file with layer names and attribute names.
        output_fc (str): Path to the output feature class, including the geodatabase.

    Raises:
        Exception: If any errors occur during processing.
    """
    try:
        # Extract geodatabase and output feature class name
        output_gdb = os.path.dirname(output_fc)
        result_fc_name = os.path.basename(output_fc)

        # If the geodatabase exists, delete it
        if arcpy.Exists(output_gdb):
            arcpy.Delete_management(output_gdb)

        # Create a new geodatabase
        arcpy.CreateFileGDB_management(os.path.dirname(output_gdb), os.path.basename(output_gdb))

        # Temporary workspace for intermediate layers
        temp_workspace = os.path.join(output_gdb, "temp")
        arcpy.CreateFeatureDataset_management(output_gdb, "temp", spatial_reference=None)

        layers_to_merge = []  # List to store processed layers
        with open(input_file, "r") as file:
            for idx, line in enumerate(file):
                line = line.strip()
                if not line or ";" not in line:
                    raise ValueError(f"Invalid line format: {line}")

                layer_path, attribute_name = line.strip().split(";")
                layer_name = f"data_{idx}"

                # Check if input layer exists
                if not arcpy.Exists(layer_path):
                    raise FileNotFoundError(f"Layer '{layer_path}' not found.")

                # Copy layer to temp workspace
                temp_layer = os.path.join(temp_workspace, layer_name)
                arcpy.CopyFeatures_management(layer_path, temp_layer)

                # Add new field T_x
                new_field = f"T_{idx}"
                arcpy.AddField_management(temp_layer, new_field, "LONG")

                # Calculate the new field based on the specified attribute
                arcpy.CalculateField_management(temp_layer, new_field, f"!{attribute_name}!", "PYTHON3")

                # Delete all fields except Shape and the new field
                fields_to_keep = ["Shape", new_field, "Shape_Length", "Shape_Area", "OBJECTID"]
                fields = [f.name for f in arcpy.ListFields(temp_layer) if f.name not in fields_to_keep]
                if fields:
                    arcpy.DeleteField_management(temp_layer, fields)

                # Add processed layer to list for merging
                layers_to_merge.append(temp_layer)

        # intersect all processed layers into the output feature class
        arcpy.Intersect_analysis(layers_to_merge, output_fc)

        # Clean up temporary layers
        arcpy.Delete_management(temp_workspace)
        print(f"Processing complete. Result saved to: {output_fc}")
        return len(layers_to_merge)

    except Exception as e:
        print(f"An error occurred: {e}")
        raise


def remove_small_pieces(in_polygon, limit, iters, n, out_polygon):

    attribs = [f"T_{i}" for i in range(n)]
    arcpy.MultipartToSinglepart_management(in_polygon, 'tmpx_p_sp')

    last_cnt = -1
    for i in range(0, iters):
        result = arcpy.GetCount_management('tmpx_p_sp')
        cnt = int(result[0])


        if cnt == last_cnt:
            break

        print(f"  patches: {cnt}")  # for information

        arcpy.MakeFeatureLayer_management('tmpx_p_sp', 'tmp_sel')
        arcpy.SelectLayerByAttribute_management('tmp_sel', 'NEW_SELECTION', 'Shape_Area < {}'.format(limit))
        arcpy.Eliminate_management('tmp_sel', 'tmpx_p', 'AREA')
        arcpy.Delete_management('tmp_sel')

        arcpy.Dissolve_management('tmpx_p', 'tmpx_p_diss', attribs)
        arcpy.MultipartToSinglepart_management('tmpx_p_diss', 'tmpx_p_sp')

        last_cnt = cnt

    arcpy.Rename_management('tmpx_p_sp', out_polygon)

    arcpy.MakeFeatureLayer_management(out_polygon, 'tmp_sel')
    arcpy.SelectLayerByAttribute_management('tmp_sel', 'NEW_SELECTION', 'Shape_Area < {}'.format(limit))
    arcpy.DeleteFeatures_management('tmp_sel')
    arcpy.Delete_management('tmp_sel')

    print(f"Processing complete. Result saved to: {out_polygon}")


def calculate_summary_field(fc, n, separator = ''):

    #arcpy.AddField_management(fc, 'CHANGE', 'TEXT', field_length=100)
    qry = separator.join(f"'+str(!T_{i}!)+'" for i in range(n))
    qry = qry[2:-2]
    arcpy.CalculateField_management(fc, 'CHANGE', qry, 'PYTHON3')
    arcpy.DeleteField_management(fc, 'ORIG_FID')


def parse_arguments():
    """
    Parses and validates command-line arguments for the ts_analyzer_area_tn.py script.

    Returns:
        dict: Dictionary of validated arguments with keys corresponding to parameter names.
    """
    parser = argparse.ArgumentParser(
        description="Analyze areas based on input vector layers and export results."
    )

    # Define required arguments
    parser.add_argument("-i", "--input_file", required=True,
                        help="Path to the input file containing vector layers.")
    parser.add_argument("-og", "--output_geodatabase", required=True,
                        help="Path to the output ESRI file geodatabase.")
    parser.add_argument("-of", "--output_feature_class", required=True,
                        help="Name of the output feature class.")

    # Define optional arguments
    parser.add_argument("-size", "--size_limit", type=int, default=5,
                        help="Minimum size of the mapping unit (default: 5).")
    parser.add_argument("-iters", "--max_iters", type=int, default=10,
                        help="Maximum number of iterations for simplification (default: 10).")
    parser.add_argument("-sep", "--separator", type=str, default="-",
                        help="Separator for the resulting column (default: '-').")

    # Parse arguments
    args = parser.parse_args()

    # Validate arguments
    if not os.path.isfile(args.input_file):
        raise ValueError(f"The input file '{args.input_file}' does not exist.")
    if not os.path.splitext(args.output_geodatabase)[1].lower() == ".gdb":
        raise ValueError(f"The output geodatabase '{args.output_geodatabase}' must have a .gdb extension.")
    if len(args.separator) != 1:
        raise ValueError("The separator must be a single character.")

    # Return arguments as a dictionary
    return {
        "input_file": args.input_file,
        "output_geodatabase": args.output_geodatabase,
        "output_feature_class": args.output_feature_class,
        "size_limit": args.size_limit,
        "max_iters": args.max_iters,
        "separator": args.separator,
    }

import arcpy
import os

def cleanup_geodatabase_except(geodatabase, keep_feature_class):
    """
    Removes all feature classes in a geodatabase except the specified one.

    Parameters:
        geodatabase (str): Path to the ESRI file geodatabase.
        keep_feature_class (str): Name of the feature class to keep.

    Raises:
        ValueError: If the specified geodatabase does not exist or is not valid.
        ValueError: If the feature class to keep does not exist in the geodatabase.
    """
    # Validate the geodatabase
    if not arcpy.Exists(geodatabase) or not geodatabase.endswith(".gdb"):
        raise ValueError(f"The geodatabase '{geodatabase}' does not exist or is not valid.")

    # Set the workspace
    arcpy.env.workspace = geodatabase

    # Check if the feature class to keep exists
    if not arcpy.Exists(os.path.join(geodatabase, keep_feature_class)):
        raise ValueError(f"The feature class '{keep_feature_class}' does not exist in the geodatabase.")

    # Get a list of all feature classes in the geodatabase
    feature_classes = arcpy.ListFeatureClasses()

    # Delete all feature classes except the one to keep
    for fc in feature_classes:
        if fc != keep_feature_class:
            arcpy.Delete_management(fc)


if __name__ == "__main__":
    try:
        result = parse_arguments()
        input_file = result["input_file"]
        output_geodatabase = result["output_geodatabase"]
        output_fc = result["output_feature_class"]
        size_limit = result["size_limit"]
        iters = result["max_iters"]
        separator = result["separator"]

        # input_file = 'S:\\changes\\classification_list.txt'
        # output_geodatabase = 'S:\\changes\\result.gdb'
        # output_fc = 'result'
        # size_limit = 5
        # iters = 10
        # separator = '-'

        arcpy.env.overwriteOutput = 1
        tmp_merged = 'tmp_data_merged_isect'
        n = process_vector_layers(input_file, os.path.join(output_geodatabase, tmp_merged))
        arcpy.env.workspace = output_geodatabase
        remove_small_pieces(tmp_merged, size_limit, iters, n, 'result')
        calculate_summary_field('result', n, separator)
        cleanup_geodatabase_except(output_geodatabase, output_fc)

        print('Finished!')

    except (FileNotFoundError, ValueError, RuntimeError) as e:
        print(e)