import arcpy
import os
import tempfile
import matplotlib.pyplot as plt
import csv
import argparse

def extract_raster_values(file_name, coordinates):
    """
    Extracts interpolated values from specific raster bands at given geographic coordinates.

    Parameters:
        file_name (str): Path to the file containing raster paths and band numbers.
        coordinates (tuple): Geographic coordinates (x, y) to extract values.

    Returns:
        dict: A dictionary where keys are raster file names (without paths) and values are the extracted raster values.

    Raises:
        FileNotFoundError: If the input file does not exist.
        ValueError: If a line in the input file is improperly formatted.
        RuntimeError: If raster processing fails.
    """
    if not os.path.isfile(file_name):
        raise FileNotFoundError(f"The file '{file_name}' does not exist.")

    # Temporary feature class to hold the point
    temp_dir = tempfile.mkdtemp()
    temp_fc = os.path.join(temp_dir, "temp_point.shp")

    # Create a temporary point feature class for the coordinates
    spatial_reference = arcpy.SpatialReference(4326)  # WGS 84
    arcpy.management.CreateFeatureclass(
        out_path=temp_dir,
        out_name="temp_point.shp",
        geometry_type="POINT",
        spatial_reference=spatial_reference
    )

    # Add the coordinate point to the feature class
    with arcpy.da.InsertCursor(temp_fc, ["SHAPE@XY"]) as cursor:
        cursor.insertRow([coordinates])

    results = {}

    with open(file_name, 'r') as file:
        for line in file:
            line = line.strip()
            if not line or "*" not in line:
                raise ValueError(f"Invalid line format: {line}")

            raster_path, band_number = line.split("*")
            raster_path = raster_path.strip()
            band_number = band_number.strip()

            if not os.path.isfile(raster_path):
                raise FileNotFoundError(f"The raster file '{raster_path}' does not exist.")

            try:
                # Build a raster layer for the specific band
                raster_layer = f"raster_band_{band_number}"
                arcpy.management.MakeRasterLayer(raster_path, raster_layer, band_index=int(band_number) - 1)

                # Extract value to the temporary point feature class
                temp_output = os.path.join(temp_dir, "output_points.shp")
                arcpy.sa.ExtractValuesToPoints(
                    in_point_features=temp_fc,
                    in_raster=raster_layer,
                    out_point_features=temp_output,
                    interpolate_values="INTERPOLATE"
                )

                # Read the extracted value
                with arcpy.da.SearchCursor(temp_output, ["RASTERVALU"]) as cursor:
                    for row in cursor:
                        value = row[0]
                        break

                # Store the result in the dictionary
                raster_name = os.path.basename(raster_path)
                results[raster_name] = value

            except Exception as e:
                raise RuntimeError(f"Error processing raster '{raster_path}': {e}")

    # Cleanup temporary files
    arcpy.management.Delete(temp_fc)
    arcpy.management.Delete(temp_output)
    return results


def create_line_chart(data_dict, output_pdf):
    """
    Creates a line chart from a dictionary and saves it as a PDF.

    Parameters:
        data_dict (dict): Dictionary where keys are labels (x-axis) and values are numbers (y-axis).
        output_pdf (str): The path to the output PDF file.

    Returns:
        None
    """
    if not data_dict:
        raise ValueError("The input dictionary is empty.")

    # Extract keys and values from the dictionary
    labels = list(data_dict.keys())
    values = list(data_dict.values())

    # Create the plot
    plt.figure(figsize=(10, 6))
    plt.plot(labels, values, color="red", marker="o", linestyle="-", linewidth=2, markersize=6)

    # Set titles and labels
    plt.xlabel("Raster", fontsize=12)
    plt.ylabel("Value", fontsize=12)
    plt.grid(True, which="both", linestyle="--", linewidth=0.5, alpha=0.7)

    # Rotate x-axis labels if needed
    plt.xticks(rotation=45, ha="right")

    # Save the plot as a PDF
    plt.tight_layout(pad=1.5)  # Adjust layout to prevent clipping
    plt.savefig(output_pdf, format="pdf")
    plt.close()


def save_dict_to_csv(data_dict, output_csv):
    """
    Saves a dictionary to a CSV file with specified headers.

    Parameters:
        data_dict (dict): Dictionary where keys are names and values are numbers.
        output_csv (str): The path to the output CSV file.

    Returns:
        None
    """
    if not data_dict:
        raise ValueError("The input dictionary is empty.")

    # Open the CSV file for writing
    with open(output_csv, mode='w', newline='', encoding='utf-8') as file:
        writer = csv.writer(file)

        # Write the header
        writer.writerow(["RASTER", "VALUE"])

        # Write each key-value pair
        for key, value in data_dict.items():
            writer.writerow([key, value])


def validate_file(file_path):
    """Check if the specified file exists."""
    if not os.path.isfile(file_path):
        raise argparse.ArgumentTypeError(f"File '{file_path}' does not exist.")
    return file_path


def validate_coords(coords):
    """Validate the format and range of geographic coordinates."""
    try:
        lat, lon = map(float, coords.split(","))
        if not (-90 <= lat <= 90 and -180 <= lon <= 180):
            raise ValueError
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Coordinates must be in the format 'latitude,longitude' with latitude in [-90, 90] and longitude in [-180, 180]."
        )
    return (lat, lon)


def parse_arguments():
    """Parse and validate command-line arguments, returning them as a dictionary."""
    parser = argparse.ArgumentParser(
        description="Time series analyzer for point data from raster files."
    )

    # Required parameters
    parser.add_argument("-i", "--input_file", required=True, type=validate_file,
                        help="Path to the file containing input raster definitions.")
    parser.add_argument("-c", "--coords", required=True, type=validate_coords,
                        help="Geographic coordinates in the format 'latitude,longitude'.")
    parser.add_argument("-o", "--output_data", required=True,
                        help="Path to the output CSV file containing raster values.")

    # Optional parameter
    parser.add_argument("-g", "--output_chart", required=False,
                        help="Path to the output PDF file for the chart.")

    args = parser.parse_args()

    # Validate output file extensions
    if not args.output_data.lower().endswith(".csv"):
        parser.error("Output data must be a CSV file with the '.csv' extension.")
    if args.output_chart and not args.output_chart.lower().endswith(".pdf"):
        parser.error("Output graph must be a PDF file with the '.pdf' extension.")

    # Return parameters as a dictionary
    return {
        "input_file": args.input_file,
        "coords": args.coords,
        "output_data": args.output_data,
        "output_chart": args.output_chart
    }


if __name__ == "__main__":
    try:
        result = parse_arguments()

        file_name = result['input_file']
        coordinates = result['coords']
        output_chart = result['output_chart']
        output_data = result['output_data']

        # file_name = "S:\\indices\\list.txt"
        # coordinates = (15.7073686, 50.7200598)
        # output_chart = "S:\\indices\\output_chart.pdf"
        # output_data = "S:\\indices\\output_data.csv"

        arcpy.env.overwriteOutput = 1

        raster_values = extract_raster_values(file_name, coordinates)
        save_dict_to_csv(raster_values, output_data)
        if output_chart:
            create_line_chart(raster_values, output_chart)
        print('Finished!')

    except (FileNotFoundError, ValueError, RuntimeError) as e:
        print(e)
